@@ -41,6 +41,7 @@ const char *fz_pdfocr_write_options_usage =
 	"\tcompression=flate: Flate compression\n"
 	"\tstrip-height=N: Strip height (default 0=fullpage)\n"
 	"\tocr-language=<lang>: OCR language (default=eng)\n"
+	"\tocr-datadir=<datadir>: OCR data path (default=rely on TESSDATA_PREFIX)\n"
 	"\n";
 
 static const char funky_font[] =
@@ -175,7 +176,11 @@ fz_parse_pdfocr_options(fz_context *ctx, fz_pdfocr_options *opts, const char *ar
 	}
 	if (fz_has_option(ctx, args, "ocr-language", &val))
 	{
-		fz_strlcpy(opts->language, val, sizeof(opts->language));
+		fz_copy_option(ctx, val, opts->language, nelem(opts->language));
+	}
+	if (fz_has_option(ctx, args, "ocr-datadir", &val))
+	{
+		fz_copy_option(ctx, val, opts->datadir, nelem(opts->datadir));
 	}
 
 	return opts;
@@ -198,6 +203,7 @@ fz_write_pixmap_as_pdfocr(fz_context *ctx, fz_output *out, const fz_pixmap *pixm
 	{
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 	}
 	fz_always(ctx)
 		fz_drop_band_writer(ctx, writer);
@@ -225,7 +231,7 @@ typedef struct pdfocr_band_writer_s
 	void *tessapi;
 	fz_pixmap *ocrbitmap;
 
-	int (*progress)(fz_context *, void *, int);
+	fz_pdfocr_progress_fn *progress;
 	void *progress_arg;
 } pdfocr_band_writer;
 
@@ -614,14 +620,15 @@ queue_word(fz_context *ctx, char_callback_data_t *cb)
 	word->len = cb->word_len;
 	memcpy(word->bbox, cb->word_bbox, 4*sizeof(float));
 	memcpy(word->chars, cb->word_chars, cb->word_len * sizeof(int));
-	word->dirn = cb->word_dirn;
 	cb->word_len = 0;
-	cb->word_dirn = 0;
 
 	line_is_v = !!(cb->line_dirn & (WORD_CONTAINS_B2T | WORD_CONTAINS_T2B));
-	word_is_v = !!(cb->line_dirn & (WORD_CONTAINS_B2T | WORD_CONTAINS_T2B));
+	word_is_v = !!(cb->word_dirn & (WORD_CONTAINS_B2T | WORD_CONTAINS_T2B));
 	line_is_h = !!(cb->line_dirn & (WORD_CONTAINS_L2R | WORD_CONTAINS_R2L));
-	word_is_h = !!(cb->line_dirn & (WORD_CONTAINS_L2R | WORD_CONTAINS_R2L));
+	word_is_h = !!(cb->word_dirn & (WORD_CONTAINS_L2R | WORD_CONTAINS_R2L));
+
+	word->dirn = cb->word_dirn;
+	cb->word_dirn = 0;
 
 	/* Can we put the new word onto the end of the existing line? */
 	if (cb->line != NULL &&
@@ -749,7 +756,7 @@ pdfocr_progress(fz_context *ctx, void *arg, int prog)
 	if (writer->progress == NULL)
 		return 0;
 
-	return writer->progress(ctx, writer->progress_arg, prog);
+	return writer->progress(ctx, writer->progress_arg, writer->pages - 1, prog);
 }
 
 static void
@@ -824,13 +831,13 @@ pdfocr_write_trailer(fz_context *ctx, fz_band_writer *writer_)
 }
 
 static void
-pdfocr_drop_band_writer(fz_context *ctx, fz_band_writer *writer_)
+pdfocr_close_band_writer(fz_context *ctx, fz_band_writer *writer_)
 {
 	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
 	fz_output *out = writer->super.out;
 	int i;
 
-	/* We actually do the trailer writing in the drop */
+	/* We actually do the trailer writing in the close */
 	if (writer->xref_max > 2)
 	{
 		int64_t t_pos;
@@ -858,13 +865,17 @@ pdfocr_drop_band_writer(fz_context *ctx, fz_band_writer *writer_)
 			fz_write_printf(ctx, out, "%010ld 00000 n \n", writer->xref[i]);
 		fz_write_printf(ctx, out, "trailer\n<</Size %d/Root 1 0 R>>\nstartxref\n%ld\n%%%%EOF\n", writer->obj_num, t_pos);
 	}
+}
 
+static void
+pdfocr_drop_band_writer(fz_context *ctx, fz_band_writer *writer_)
+{
+	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
 	fz_free(ctx, writer->stripbuf);
 	fz_free(ctx, writer->compbuf);
 	fz_free(ctx, writer->page_obj);
 	fz_free(ctx, writer->xref);
 	fz_drop_pixmap(ctx, writer->ocrbitmap);
-
 	ocr_fin(ctx, writer->tessapi);
 }
 #endif
@@ -879,6 +890,7 @@ fz_band_writer *fz_new_pdfocr_band_writer(fz_context *ctx, fz_output *out, const
 	writer->super.header = pdfocr_write_header;
 	writer->super.band = pdfocr_write_band;
 	writer->super.trailer = pdfocr_write_trailer;
+	writer->super.close = pdfocr_close_band_writer;
 	writer->super.drop = pdfocr_drop_band_writer;
 
 	if (options)
@@ -900,7 +912,7 @@ fz_band_writer *fz_new_pdfocr_band_writer(fz_context *ctx, fz_output *out, const
 
 	fz_try(ctx)
 	{
-		writer->tessapi = ocr_init(ctx, writer->options.language);
+		writer->tessapi = ocr_init(ctx, writer->options.language, writer->options.datadir);
 	}
 	fz_catch(ctx)
 	{
@@ -913,7 +925,7 @@ fz_band_writer *fz_new_pdfocr_band_writer(fz_context *ctx, fz_output *out, const
 }
 
 void
-fz_pdfocr_band_writer_set_progress(fz_context *ctx, fz_band_writer *writer_, int (*progress)(fz_context *, void *, int), void *progress_arg)
+fz_pdfocr_band_writer_set_progress(fz_context *ctx, fz_band_writer *writer_, fz_pdfocr_progress_fn *progress, void *progress_arg)
 {
 #ifdef OCR_DISABLED
 	fz_throw(ctx, FZ_ERROR_GENERIC, "No OCR support in this build");
@@ -996,9 +1008,7 @@ pdfocr_close_writer(fz_context *ctx, fz_document_writer *wri_)
 {
 	fz_pdfocr_writer *wri = (fz_pdfocr_writer*)wri_;
 
-	fz_drop_band_writer(ctx, wri->bander);
-	wri->bander = NULL;
-
+	fz_close_band_writer(ctx, wri->bander);
 	fz_close_output(ctx, wri->out);
 }
 
@@ -1054,7 +1064,7 @@ fz_new_pdfocr_writer(fz_context *ctx, const char *path, const char *options)
 }
 
 void
-fz_pdfocr_writer_set_progress(fz_context *ctx, fz_document_writer *writer, int (*progress)(fz_context *, void *, int), void *progress_arg)
+fz_pdfocr_writer_set_progress(fz_context *ctx, fz_document_writer *writer, fz_pdfocr_progress_fn *progress, void *progress_arg)
 {
 #ifdef OCR_DISABLED
 	fz_throw(ctx, FZ_ERROR_GENERIC, "No OCR support in this build");
